//
//  iTeleportAppDelegate.m
//  iTeleport
//
//  Created by Samuco on 1/07/10.
//  Copyright 2010 Samuco. All rights reserved.
//

#import "iTeleportAppDelegate.h"
#import "VMRegion.h"
#import "SearchContext.h"
#import "Variable.h"

#import <SecurityFoundation/SFAuthorization.h>
#import <Security/AuthorizationTags.h>
#import "unistd.h"

void authMe(char * FullPathToMe)
{
	// get authorization as root
	
	OSStatus myStatus;
	
	// set up Authorization Item
	AuthorizationItem myItems[1];
	myItems[0].name = kAuthorizationRightExecute;
	myItems[0].valueLength = 0;
	myItems[0].value = NULL;
	myItems[0].flags = 0;
	
	// Set up Authorization Rights
	AuthorizationRights myRights;
	myRights.count = sizeof (myItems) / sizeof (myItems[0]);
	myRights.items = myItems;
	
	// set up Authorization Flags
	AuthorizationFlags myFlags;
	myFlags =
	kAuthorizationFlagDefaults |
	kAuthorizationFlagInteractionAllowed |
	kAuthorizationFlagExtendRights;
	
	// Create an Authorization Ref using Objects above. NOTE: Login bod comes up with this call.
	AuthorizationRef myAuthorizationRef;
	myStatus = AuthorizationCreate (&myRights, kAuthorizationEmptyEnvironment, myFlags, &myAuthorizationRef);
	
	if (myStatus == errAuthorizationSuccess)
	{
		// prepare communication path - used to signal that process is loaded
		FILE *myCommunicationsPipe = NULL;
		char myReadBuffer[] = " ";
		
		// run this app in GOD mode by passing authorization ref and comm pipe (asynchoronous call to external application)
		myStatus = AuthorizationExecuteWithPrivileges(myAuthorizationRef,FullPathToMe,kAuthorizationFlagDefaults,nil,&myCommunicationsPipe);
		
		// external app is running asynchronously - it will send to stdout when loaded
		if (myStatus == errAuthorizationSuccess)
		{
			read (fileno (myCommunicationsPipe), myReadBuffer, sizeof (myReadBuffer));
			fclose(myCommunicationsPipe);
		}
		
		// release authorization reference
		myStatus = AuthorizationFree (myAuthorizationRef, kAuthorizationFlagDestroyRights);
	}
}

bool checkExecutablePermissions(void)
{
	NSDictionary	*applicationAttributes = [[NSFileManager defaultManager] fileAttributesAtPath:[[NSBundle mainBundle] executablePath] traverseLink: YES];
	
	// We expect 2755 as octal (1517 as decimal, -rwxr-sr-x as extended notation)
	return ([applicationAttributes filePosixPermissions] == 1517 && [[applicationAttributes fileGroupOwnerAccountName] isEqualToString: @"procmod"]);
}

bool amIWorthy(void)
{
	// running as root?
	AuthorizationRef myAuthRef;
	OSStatus stat = AuthorizationCopyPrivilegedReference(&myAuthRef,kAuthorizationFlagDefaults);
	
	return stat == errAuthorizationSuccess || checkExecutablePermissions();
}



@implementation iTeleportAppDelegate

@synthesize window;

-(void)speak:(NSString*)words
{
	NSString *voi = @"Vicki";
	
	NSAppleScript *script = [[NSAppleScript alloc] initWithSource:[NSString stringWithFormat:@"say \"%@\" using \"%@\"", words, voi]];
	[script executeAndReturnError:nil];
}

- (void)speechRecognizer:(NSSpeechRecognizer *)sender didRecognizeCommand:(id)command
{
	
	NSString *com = (NSString*)command;
	if ([com isEqualToString:@"System"])
	{
		NSBeep();
		//[self speak:@"State location?"];
		
		[voice setCommands:commands];
	}
	else {
		//Work out 
		NSString *loc = [locations objectAtIndex:[commands indexOfObject:com]];
		NSArray *pieces = [loc componentsSeparatedByString:@" "];
		
		NSLog(loc);
		
		float x = [[pieces objectAtIndex:0] floatValue];
		float y = [[pieces objectAtIndex:1] floatValue];
		float z = [[pieces objectAtIndex:2] floatValue];
		
		
		const int hostObjectIDAddress = 0x4BD7B002 + 0x200 * 0;
		const int firstTableObjectArrayAddress = 0x4BB206EC;
		const int tableObjectArraySize = 12;
		const unsigned short invalidHostObjectID = 0xFFFF;
		const int offsetToObjectArrayTablePointer = 0x8;
		
		unsigned short hostObjectID;
		vm_size_t size = sizeof(short);
		VMReadBytes(haloProcessID, hostObjectIDAddress, &hostObjectID, &size);
		hostObjectID = CFSwapInt16BigToHost(hostObjectID);
		
		
		if (hostObjectID != 0 && hostObjectID != invalidHostObjectID)
		{
			unsigned int haloObjectPointerAddress = firstTableObjectArrayAddress + hostObjectID * tableObjectArraySize + offsetToObjectArrayTablePointer;
			
			int haloObjectPointer;
			size = sizeof(int);
			VMReadBytes(haloProcessID, haloObjectPointerAddress, &haloObjectPointer, &size);
			haloObjectPointer = CFSwapInt32BigToHost(haloObjectPointer);
			
			
			int vehicle_mode = 1;
			
			if (vehicle_mode)
			{
				
				//VEHICLE
				const int offsetToPlayerVehicle = haloObjectPointer + 2 + 2 + 88 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 4 + 84 + 4 + 4 + 48 + 2 + 2 + 2;
				
				unsigned short haloObjectVehi;
				vm_size_t size = sizeof(short);
				VMReadBytes(haloProcessID, offsetToPlayerVehicle, &haloObjectVehi, &size);
				haloObjectVehi = CFSwapInt16BigToHost(haloObjectVehi);
				
				
				unsigned int tablez_address = firstTableObjectArrayAddress + haloObjectVehi * tableObjectArraySize + offsetToObjectArrayTablePointer;
				//const int tablez_address = haloObjectVehi;//firstTableObjectArrayAddress + (haloObjectVehi * tableObjectArraySize);
				unsigned int haloObjectPointerAddress2 = tablez_address;
				
				int	haloObjectPointer2;
				size = sizeof(int);
				VMReadBytes(haloProcessID, haloObjectPointerAddress2, &haloObjectPointer2, &size);
				haloObjectPointer2 = CFSwapInt32BigToHost(haloObjectPointer2);
				
				haloObjectPointer = haloObjectPointer2;
				
			}
			
			//END VEHICLE
			
			const int offsetToPlayerXCoordinate = 0x5C;
			const int offsetToPlayerYCoordinate = 0x5C + 0x4;
			const int offsetToPlayerZCoordinate = 0x5C + 0x4 + 0x4;
			
			float new_value = [[pieces objectAtIndex:0] floatValue];
			int *valueP = (int *)&new_value;
			*valueP = CFSwapInt32HostToBig(*((int *)&new_value));
			
			VMWriteBytes(haloProcessID, haloObjectPointer + offsetToPlayerXCoordinate, &new_value, sizeof(float));
			 
			new_value = [[pieces objectAtIndex:1] floatValue];
			valueP = (int *)&new_value;
			*valueP = CFSwapInt32HostToBig(*((int *)&new_value));
			
			VMWriteBytes(haloProcessID, haloObjectPointer + offsetToPlayerYCoordinate, &new_value, sizeof(float));
			
			new_value = [[pieces objectAtIndex:2] floatValue];
			valueP = (int *)&new_value;
			*valueP = CFSwapInt32HostToBig(*((int *)&new_value));
			
			VMWriteBytes(haloProcessID, haloObjectPointer + offsetToPlayerZCoordinate, &new_value, sizeof(float));
			
			NSBeep();
		}
		else {
			[self speak:@"You're dead."];
		}
		
		//[voice setCommands:[[NSMutableArray alloc] initWithObjects:@"System", nil]];

	}

}

-(void)writeFloat:(float)value to:(int)address
{
	// Kill the host!
	float new_value = value;
	
	int *valueP = (int *)&new_value;
	*valueP = CFSwapInt32HostToBig(*((int *)&new_value));
	
	VMWriteBytes(haloProcessID, address, &new_value, sizeof(float));
}

- (void)applicationDidFinishLaunching:(NSNotification *)aNotification {
	
	
	//Set the process ID
	haloProcessID = 0;
	
	BOOL needsAuthorizationCheck = YES;
	
#ifdef __ppc__
	// PPC machines whose operating system is below leopard do not need authorization
	SInt32 osxMinorVersion;
	Gestalt(gestaltSystemVersionMinor, &osxMinorVersion);
	if (osxMinorVersion < 5)
	{
		needsAuthorizationCheck = NO;
	}
#endif
	
	if (needsAuthorizationCheck)
	{
		if (amIWorthy())
		{
			// when the app restarts from asking the user his adminstator's password, the app isn't activated for some reason, so we'll activate it
			//[NSApp activateIgnoringOtherApps:YES];
			
#ifndef _DEBUG
			// if we don't print this, the old instance of the app will stll be alive. this behavior is wanted in debug mode because xcode's debugger will remain opened
			printf("Don't forget to flush! ;-) "); // signal back to close caller
#endif
			fflush(stdout);
		
	
			
			NSDictionary *applicationDictionary;
			NSEnumerator *launchedApplicationsEnumerator = [[[NSWorkspace sharedWorkspace] launchedApplications] objectEnumerator];
			while ((applicationDictionary = [launchedApplicationsEnumerator nextObject]))
			{
				[self assignHaloProcessIDFromApplicationDictionary:applicationDictionary];
			}
			
			[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self
																   selector:@selector(applicationDidLaunch:)
																	   name:NSWorkspaceDidLaunchApplicationNotification
																	 object:nil];
			
			[[[NSWorkspace sharedWorkspace] notificationCenter] addObserver:self
																   selector:@selector(applicationDidTerminate:)
																	   name:NSWorkspaceDidTerminateApplicationNotification
																	 object:nil];
			
	// Insert code here to initialize your application 
	locations = [[NSMutableArray alloc] init];
	commands = [[NSMutableArray alloc] init];

	[commands addObject:@"Cave one"];
	[locations addObject:@"100.634048462 -111.866004944 4.24200153351"];
	
	[commands addObject:@"Red base"];
	[locations addObject:@"95.470916748 -160.95123291 1.70348572731"];
	
	[commands addObject:@"Blu base"];
	[locations addObject:@"40.1523208618 -77.7251815796 1.70348572731"];
	
	[commands addObject:@"Red tree"];
	[locations addObject:@"79.6294784546 -149.14112854 0.136964842677"];
	
	[commands addObject:@"Blu tree"];
	[locations addObject:@"37.983581543 -96.7820968628 -0.125067532063"];
	
	[commands addObject:@"middle"];
	[locations addObject:@"64.8017959595 -120.15247345 0.141527757049"];
	
	[commands addObject:@"Red cliff"];
	[locations addObject:@"71.5347900391 -87.5424499512 5.7582359314"];
	
	[commands addObject:@"Blu cliff"];
	[locations addObject:@"61.8856468201 -154.115875244 6.16351461411"];
	
	[commands addObject:@"Cliff one"];
	[locations addObject:@"74.7244873047 -89.7450637817 23.4769897461"];
	
	[commands addObject:@"Cliff two"];
	[locations addObject:@"51.888874054 -156.127838135 21.836566925"];
	
	[commands addObject:@"Behind red"];
	[locations addObject:@"61.7425575256 -179.37008667 4.55214309692"];
	
	[commands addObject:@"Behind blu"];
	[locations addObject:@"72.428276062 -62.4241981506 4.39192438126"];
	
	[commands addObject:@"Cliff three"];
	[locations addObject:@"90.993057251 -116.21497345 17.1609249115"];
			
	[commands addObject:@"Flame"];
	[locations addObject:@"104.05670166 -109.288497925 1.71249318123"];
			
	[commands addObject:@"Blu flag"];
	[locations addObject:@"40.2135810852 -79.1399230957 -0.286985069513"];
			
	[commands addObject:@"Red flag"];
	[locations addObject:@"95.7080612183 -159.422393799 -0.286985069513"];
			
	voice = [[NSSpeechRecognizer alloc] init];
	
	[voice setDelegate:self];
	[voice setBlocksOtherRecognizers:YES];
	[voice setListensInForegroundOnly:NO];
	
	[voice startListening];
	[voice setCommands:commands];
			
		}
		else
		{
			extern char *gExecutablePath;
			authMe(gExecutablePath);
			[[NSApplication sharedApplication] terminate:nil];
		}
	}
}

- (void)assignHaloProcessIDFromApplicationDictionary:(NSDictionary *)applicationDictionary
{
	if ([[applicationDictionary objectForKey:@"NSApplicationName"] rangeOfString:@"Halo Demo" options:NSCaseInsensitiveSearch].location != NSNotFound)
	{
		haloProcessID = [[applicationDictionary objectForKey:@"NSApplicationProcessIdentifier"] intValue];
		NSLog(@"PID %d", haloProcessID);
	}
}

- (void)applicationDidLaunch:(NSNotification *)notification
{
	[self assignHaloProcessIDFromApplicationDictionary:[notification userInfo]];
}

- (void)applicationDidTerminate:(NSNotification *)notification
{
	if ([[[notification userInfo] objectForKey:@"NSApplicationName"] rangeOfString:@"Halo Demo" options:NSCaseInsensitiveSearch].location != NSNotFound)
	{
		haloProcessID = 0;
	}
}

@end
